/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <libintl.h>

#include "sambascanner.h"
#include "String.h"
#include "IP_Range.h"
#include "ConfigurationFile.h"

extern struct Options_Structure Options;

static const char *KnownHost_Callback(const char *Key, const char *Value);
static const char *IPRange_Callback(const char *Key, const char *Value);
static const char *UserPass_Callback(const char *Key, const char *Value);

struct ConfigFileDirective Directives[] = {
	{ "General", "Workgroup", DIRT_STRING, { Options.Workgroup } },
	{ "General", "KnownHost", DIRT_CALLBACK, { KnownHost_Callback } },
	{ "General", "IPRange", DIRT_CALLBACK, { IPRange_Callback } },
	{ "General", "TimingMultiply", DIRT_DOUBLE, { &Options.Timings }, TIMEOUT_MIN, TIMEOUT_MAX },
	{ "General", "ThreadCount", DIRT_INT, { &Options.ThreadCount }, 1, 100 },
	{ "OutFiles", "Filelist", DIRT_STRING, { Options.Outfile } },
	{ "OutFiles", "Hostlist", DIRT_STRING, { Options.HostlistFile } },
	{ "OutFiles", "Sharelist", DIRT_STRING, { Options.SharelistFile } },
	{ "ShareAccess", "UserPass", DIRT_CALLBACK, { UserPass_Callback } },
	{ "ShareAccess", "DefaultPasswords", DIRT_INT, { &Options.TryPasswords } },
	{ "ShareAccess", "ShowPrintShares", DIRT_INT, { &Options.Show_PrintShares } },
	{ "ShareAccess", "ConnectByIP", DIRT_INT, { &Options.Connect_By_IP } },
	{ "ShareAccess", "UseMasterBrowserHints", DIRT_INT, { &Options.Secondary_Lookup } },
	{ "FileLocations", "SmbClient", DIRT_STRING, { Options.SmbClient_Executable } },
	{ "FileLocations", "NmbLookup", DIRT_STRING, { Options.NmbLookup_Executable } },
	{ "FileLocations", "SmbLister", DIRT_STRING, { Options.SmbLister_Executable } },
	{ "Debug", "DebugLevel", DIRT_INT, { &Options.DebugLvl }, 0, 6 },
	{ "Debug", "PreserveTempFiles", DIRT_INT, { &Options.PreserveTemp } },
	{ }
};

static const char *KnownHost_Callback(const char *Key, const char *Value) {
	Add_Hostname(Options.Hostlist, Value);
	return NULL;
}

static const char *IPRange_Callback(const char *Key, const char *Value) {
	int Type;
	Type = New_IP_Pattern(Value);
	if (Type) {
		return _("Invalid IP pattern");
	}
	return NULL;
}

static const char *UserPass_Callback(const char *Key, const char *Value) {
	if (strchr(Value, ',') == NULL) {
		return _("No comma supplied seperating username from password");
	}

	Add_String_Unique(&Options.UserPass, Value);

	return NULL;
}

static void ProcessLine(char *Line, int LineNumber, int ShowErrors) {
	static char CurrentSection[32];
	int Length;

	Length = strlen(Line);

	if (Length == 0) return;
	if (Line[0] == ';') return;

	if ((Line[0] == '[') && (Line[Length - 1] == ']')) {
		/* New section */
		if (Length > sizeof(CurrentSection) + 2) {
			/* Section name too long */
			if (ShowErrors) message(2, _("Section name greater than %d characters in line %d, line ignored.\n"), (int)sizeof(CurrentSection), LineNumber);
			return;
		}
		Line[Length - 1] = 0;
		strcpy(CurrentSection, Line + 1);
		return;
	}

	{
		char *Token1, *Token2;
		char *SavePtr;
		Token1 = strtok_r(Line, "=", &SavePtr);
		if (Token1 == NULL) {
			if (ShowErrors) message(2, _("No data in line %d, line ignored.\n"), LineNumber);
			return;
		}
		Remove_Trailing_Whitespaces(Token1);

		Token2 = strtok_r(NULL, "=", &SavePtr);
		if (Token2 == NULL) {
			if (ShowErrors) message(2, _("No data in line %d after key '%s', line ignored.\n"), LineNumber, Token1);
			return;
		}
		Remove_Leading_Whitespaces(Token2);

		message(5, _("Configuration file line #%d: '%s' = '%s'\n"), LineNumber, Token1, Token2);
		{
			int i;
			i = -1;
			while (1) {
				i++;
				if (Directives[i].Section == NULL) break;						/* No more sections available */
				if (strcmp(Directives[i].Section, CurrentSection)) continue;	/* Section doesn't match */
				if (strcmp(Directives[i].Name, Token1)) continue;				/* Key values doesn't match */
				switch (Directives[i].Type) {
					case DIRT_INT: {
						int Value;
						if (!strcmp(Token2, "true")) Value = 1;
							else if (!strcmp(Token2, "false")) Value = 0;
							else Value = atoi(Token2);
						if (Directives[i].Min != Directives[i].Max) {
							if ((Value < Directives[i].Min) || (Value > Directives[i].Max)) {
								if (ShowErrors) message(3, _("Integer value %d was out of range (%d-%d) in line %d, line ignored.\n"), Value, Directives[i].Min, Directives[i].Max, LineNumber);
								return;
							}
						}
						*Directives[i].Location.Int = Value;
						return;
						break;
					}
					case DIRT_DOUBLE: {
						double Value = atof(Token2);
						if (Directives[i].Min != Directives[i].Max) {
							if ((Value < Directives[i].Min) || (Value > Directives[i].Max)) {
								if (ShowErrors) message(3, _("Double value %.1f was out of range (%d-%d) in line %d, line ignored.\n"), Value, Directives[i].Min, Directives[i].Max, LineNumber);
								return;
							}
						}
						*Directives[i].Location.Double = Value;
						return;
						break;
					}
					case DIRT_STRING: {
						if (Directives[i].Min != Directives[i].Max) {
							int Length = strlen(Token2);
							if (((Length < Directives[i].Min) || (Length > Directives[i].Max))) {
								if (ShowErrors) message(3, _("String length (%d) was out of range (%d-%d) in line %d, line ignored.\n"), Length, Directives[i].Min, Directives[i].Max, LineNumber);
								return;
							}
						}
						strcpy(Directives[i].Location.String, Token2);
						return;
						break;
					}
					case DIRT_CALLBACK: {
						char *Error;
						Error = Directives[i].Location.Callback(Token1, Token2);
						if (Error) {
							if (ShowErrors) message(3, _("Error in line %d of configuration file: %s\n"), LineNumber, Error);
						}
						return;
						break;
					}
				}

			}

			/* All directives through, but no match found */
			/* Check if the directive is found in another section */
			{
				int First = 1;
				i = -1;
				while (Directives[++i].Section != NULL) {
					if (!strcmp(Directives[i].Name, Token1)) {
						if (First) {
							First = 0;
							message(3, _("No match for directive \"%s\", section \"%s\" in line number %d. Did you maybe mean section \"%s\""), Token1, CurrentSection, LineNumber, Directives[i].Section);
						} else {
							message(3, _(" or \"%s\""), Directives[i].Section);
						}
					}
				}
				if (!First) {
					message(3, _("?\n"));
				} else {
					message(3, _("No match for directive \"%s\", section \"%s\" in line number %d.\n"), Token1, CurrentSection, LineNumber);
				}
			}
	
		}
	}
}

enum ConfigReturnType ScanFile(const char *Filename, int ShowErrors) {
	FILE *f;
	f = fopen(Filename, "r");
	if (!f) {
		if (ShowErrors) message(2, _("Cannot open configuration file \"%s\": %s\n"), Filename, strerror(errno));
		return CONFIG_ERROR;
	}

	message(3, _("Parsing configuration file \"%s\".\n"), Filename);
	{
		int LineNumber = 0;
		char LineBuffer[BUFSIZE_SHORT];
		while (fgets(LineBuffer, sizeof(LineBuffer), f) != NULL) {
			LineNumber++;
			/* Is the line terminated by a NL? */
			int l;
			l = strlen(LineBuffer);
			if (LineBuffer[l - 1] != '\n') {
				/* Line too long */
				int c;
				while (((c = getc(f)) != EOF) && (c != '\n'));
				message(3, _("Line number %d of configuration file \"%s\" was either longer than %d characters or not terminated by newline - ignoring line.\n"), LineNumber, Filename, (int)sizeof(LineBuffer));
			}

			LineBuffer[l - 1] = 0;
			ProcessLine(LineBuffer, LineNumber, ShowErrors);
		}
	}

	fclose(f);
	return CONFIG_OK;
}

void ScanDefaultConfigurationFile() {
	char Buffer[BUFSIZE_SHORT];
	char *HomeDir;
	HomeDir = getenv("HOME");
	if (!HomeDir) {
		message(3, _("Unable to determine home directory of user, not scanning \"~/%s\" configuration file.\n"), DEFAULT_CONFIGFILE);
		return;
	}
	Buffer[sizeof(Buffer) - 1] = 0;
	strncpy(Buffer, HomeDir, sizeof(Buffer) - 1);

	if (strlen(Buffer) + 1 + strlen(DEFAULT_CONFIGFILE) + 1 > sizeof(Buffer)) {
		message(3, _("Unable to determine full filename of \"~/%s\", buffer overflow.\n"), DEFAULT_CONFIGFILE);
		return;
	}
	strcat(Buffer, "/" DEFAULT_CONFIGFILE);
	ScanFile(Buffer, 0);
}

