/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <libintl.h>

#include "sambascanner.h"
#include "IP_Range.h"

static int PatternCount = 0;
static struct Pattern **PatternMem = NULL;
extern struct Options_Structure Options;

int Have_Patterns() {
	return (PatternCount > 0);
}

void IP_To_String(ipv4addr_t IP, char *Buffer, int Size) {
	snprintf(Buffer, Size, "%d.%d.%d.%d", IP_ADDRESS(IP));
}

static void Get_Next_IP_Subnet(struct Pattern_Subnet *Pattern, struct IP_Address *Result) {
	Result->IP = Pattern->IP;
	Pattern->IP++;
	Result->Done = (Pattern->IP > Pattern->Last);
}

static void Get_Next_IP_Range(struct Pattern_Range *Pattern, struct IP_Address *Result) {
	int i;
	Result->IP = Pattern->IP;
	Result->Done = 0;
	for (i = 0; i < 4; i++) {
		int BitPattern;
		unsigned char Start, End, Current;
		BitPattern = (0xff << (8 * i));
		Start = (Pattern->Start_IP & BitPattern) >> (8 * i);
		End = (Pattern->End_IP & BitPattern) >> (8 * i);
		Current = (Pattern->IP & BitPattern) >> (8 * i);

		Current++;
		if (Current > End) {
			Current = Start;
		}
		
		Pattern->IP &= ~(BitPattern);
		Pattern->IP |= Current << (8 * i);

		if (Current != Start) return;
	}
	Result->Done = 1;
}

static void Get_Next_IP_Address(struct Pattern_Address *Pattern, struct IP_Address *Result) {
	memcpy(&(Result->IP), &(Pattern->IP), 4);
	Result->Done = 1;
}

static void Get_Next_IP_Pattern(struct Pattern *Pattern, struct IP_Address *Result) {
	if (Pattern->Type == PATTERN_SUBNET) Get_Next_IP_Subnet(Pattern->Pattern_Subnet, Result);
		else if (Pattern->Type == PATTERN_RANGE) Get_Next_IP_Range(Pattern->Pattern_Range, Result);
		else Get_Next_IP_Address(Pattern->Pattern_Address, Result);
}

int Get_Next_IP(ipv4addr_t *IP_Address) {
	static int CurrentPatternNumber = 0;
	struct IP_Address Result;
	if (CurrentPatternNumber == PatternCount) {
		/* No more patterns ! */
		(*IP_Address) = 0x00000000;
		return 0;
	} 

	Get_Next_IP_Pattern(PatternMem[CurrentPatternNumber], &Result);
	memcpy(IP_Address, &(Result.IP), sizeof(ipv4addr_t));
	if (Result.Done) CurrentPatternNumber++;
	return 1;
}

static struct Pattern_Address *Parse_Pattern_Address(const char *Pattern) {
	static struct Pattern_Address InternalPattern;
	int i;
	char Buffer[BUFSIZE_SHORT];
	char *Token;
	char *Tokenizer_SavePtr;
	const char *Delimiter = ".";
	strncpy(Buffer, Pattern, sizeof(Buffer));
	InternalPattern.IP = 0x00000000;
	
	InternalPattern.Valid = 1;
	Token = (char*)strtok_r(Buffer, Delimiter, &Tokenizer_SavePtr);
	for (i = 0; i < 4; i++) {
		int Value = atoi(Token);
		InternalPattern.IP <<= 8;
		if ((Value < 0) || (Value > 255)) {
			InternalPattern.Valid = 0;
			break;
		} else {
			InternalPattern.IP |= Value;
		}
		Token = (char*)strtok_r(NULL, Delimiter, &Tokenizer_SavePtr);
	}
/*	printf("Parse Address: "); Display_IP(InternalPattern.IP); printf("\n");*/
	return &InternalPattern;
}

static struct Pattern_Subnet *Parse_Pattern_Subnet(const char *Pattern) {
	static struct Pattern_Subnet InternalPattern;
	struct Pattern_Address *Address;
	char *Token;
	char *Tokenizer_SavePtr;
	char Buffer[BUFSIZE_SHORT];

	InternalPattern.Valid = 1;

	strncpy(Buffer, Pattern, sizeof(Buffer));
	Token = (char*)strtok_r(Buffer, "/", &Tokenizer_SavePtr);
	Address = Parse_Pattern_Address(Token);
	if (Address->Valid == 0) {
		InternalPattern.Valid = 0;
	} else {
		int Subnet;
		InternalPattern.IP = Address->IP;
		Token = (char*)strtok_r(NULL, "/", &Tokenizer_SavePtr);

		/* This is now the subnet, e.g. '24' */
		Subnet = atoi(Token);
		if ((Subnet < 0) || (Subnet > 30)) {
			InternalPattern.Valid = 0;
		} else {
			ipv4addr_t SubnetMask;
			ipv4addr_t OldIP = InternalPattern.IP;
			SubnetMask = 0xffffffff << (32 - Subnet);
			OldIP = InternalPattern.IP;
			InternalPattern.IP &= SubnetMask;
			if (OldIP != InternalPattern.IP) {
				fprintf(stderr, _("Warning: Given IP address %d.%d.%d.%d was truncated to %d.%d.%d.%d due to subnet mask of %d.%d.%d.%d\n"), IP_ADDRESS(OldIP), IP_ADDRESS(InternalPattern.IP), IP_ADDRESS(SubnetMask));
			}
			InternalPattern.Last = (1 << (32 - Subnet)) + InternalPattern.IP - 2;
			InternalPattern.IP++;
		}
	}

	return &InternalPattern;
}

static struct Pattern_Range *Parse_Pattern_Range(const char *Pattern) {
	static struct Pattern_Range InternalPattern;
	int i;
	char Buffer[BUFSIZE_SHORT];
	char *Token;
	char *Tokenizer_SavePtr;
	const char *Delimiter = ".";
	strncpy(Buffer, Pattern, sizeof(Buffer));
	InternalPattern.Start_IP = 0x00000000;
	InternalPattern.End_IP = 0x00000000;
	
	InternalPattern.Valid = 1;
	Token = (char*)strtok_r(Buffer, Delimiter, &Tokenizer_SavePtr);
	for (i = 0; i < 4; i++) {
		int Value1, Value2;
		char Buffer[BUFSIZE_SHORT];
		char *InnerToken;
		char *Tokenizer2_SavePtr;
		
		/* Is Token a number or a range? */
		strncpy(Buffer, Token, sizeof(Buffer));
	
		InnerToken = (char*)strtok_r(Buffer, "-", &Tokenizer2_SavePtr);
		Value1 = atoi(InnerToken);

		InnerToken = (char*)strtok_r(NULL, "-", &Tokenizer2_SavePtr);
		if (InnerToken == NULL) {
			Value2 = Value1;
		} else {
			Value2 = atoi(InnerToken);
		}

		if ((Value1 < 0) || (Value1 > 255) || (Value2 < 0) || (Value2 > 255)) {
			InternalPattern.Valid = 0;
			break;
		} else {
			if (Value1 > Value2) {
				int Temp;
				Temp = Value1;
				Value1 = Value2;
				Value2 = Temp;
			}
		}
		
		InternalPattern.Start_IP <<= 8;
		InternalPattern.Start_IP |= Value1;
		InternalPattern.End_IP <<= 8;
		InternalPattern.End_IP |= Value2;
		
		Token = (char*)strtok_r(NULL, Delimiter, &Tokenizer_SavePtr);
	}

	/*
	printf("Parse Range Start: "); Display_IP(InternalPattern.Start_IP); printf("\n");
	printf("Parse Range End  : "); Display_IP(InternalPattern.End_IP); printf("\n");
	*/
	InternalPattern.IP = InternalPattern.Start_IP;
	return &InternalPattern;
}

struct Pattern *Parse_Pattern(const char *Pattern, int Type) {
	static struct Pattern InternalPattern;
	InternalPattern.Type = Type;
	if (Type == PATTERN_SUBNET) {
		InternalPattern.Pattern_Subnet = Parse_Pattern_Subnet(Pattern);
		InternalPattern.Valid = InternalPattern.Pattern_Subnet->Valid;
	} else if (Type == PATTERN_RANGE) {
		InternalPattern.Pattern_Range = Parse_Pattern_Range(Pattern);
		InternalPattern.Valid = InternalPattern.Pattern_Range->Valid;
	} else {
		InternalPattern.Pattern_Address = Parse_Pattern_Address(Pattern);
		InternalPattern.Valid = InternalPattern.Pattern_Address->Valid;
	}
	return &InternalPattern;
}

int IP_Pattern_Valid(const char *Pattern) {
	char Buffer[BUFSIZE_SHORT];
	int Type = 0;

	/* Every valid pattern has exactly three dots and always something inbetween (i.e. at least one character) */
	{
		int i;
		char *Token;
		const char *Delimiter = ".";
		char *Tokenizer_SavePtr;
		strncpy(Buffer, Pattern, sizeof(Buffer));
		i = 0;
		Token = (char*)strtok_r(Buffer, Delimiter, &Tokenizer_SavePtr);
		while ((Token != NULL) && (strlen(Token) >= 1) && (i <= 3)) {
			Token = (char*)strtok_r(NULL, Delimiter, &Tokenizer_SavePtr);
			i++;
		}

		if ((i != 4) || (Token != NULL)) return 0;
	}

	/* There is no trailing dot */
	if (Pattern[strlen(Pattern) - 1] == '.') return 0;

	/* A valid pattern has only either up to three dashes or a single slash */
	{
		int Slashes = 0;
		int Dashes = 0;
		int i;
		for (i = 0; i < strlen(Pattern); i++) {
			if (Pattern[i] == '/') Slashes++;
				else if (Pattern[i] == '-') Dashes++;
		}
		if ((Slashes > 0) && (Dashes > 0)) return 0;
		if (Slashes > 0) Type = PATTERN_SUBNET;
			else if (Dashes > 0) Type = PATTERN_RANGE;
			else Type = PATTERN_IP;
	}

	return Type;
}

struct Pattern* CopyPattern(struct Pattern *Old) {
	struct Pattern *New;
	New = (struct Pattern*)malloc(sizeof(struct Pattern));
	memcpy(New, Old, sizeof(struct Pattern));
	if (New->Type == PATTERN_SUBNET) {
		New->Pattern_Subnet = (struct Pattern_Subnet*)malloc(sizeof(struct Pattern_Subnet));
		memcpy(New->Pattern_Subnet, Old->Pattern_Subnet, sizeof(struct Pattern_Subnet));
	} else if (New->Type == PATTERN_RANGE) {
		New->Pattern_Range = (struct Pattern_Range*)malloc(sizeof(struct Pattern_Range));
		memcpy(New->Pattern_Range, Old->Pattern_Range, sizeof(struct Pattern_Range));
	} else {
		New->Pattern_Address = (struct Pattern_Address*)malloc(sizeof(struct Pattern_Address));
		memcpy(New->Pattern_Address, Old->Pattern_Address, sizeof(struct Pattern_Address));
	}
	return New;
}

static void Free_IP_Pattern(struct Pattern *Pattern) {
	if (Pattern->Type == PATTERN_SUBNET) {
		free(Pattern->Pattern_Subnet);
	} else  if (Pattern->Type == PATTERN_RANGE) {
		free(Pattern->Pattern_Range);
	} else {
		free(Pattern->Pattern_Address);
	}
	free(Pattern);
}

void Free_All_Patterns() {
	int i;
	for (i = 0; i < PatternCount; i++) {
		Free_IP_Pattern(PatternMem[i]);
	}
	free(PatternMem);
}

const char* Pattern_Type_To_String(int Type) {
	switch (Type) {
		case PATTERN_SUBNET: return "Subnet";
		case PATTERN_RANGE : return "Range";
		case PATTERN_IP    : return "IP Address";
	}
	return "Unknown";
}

int New_IP_Pattern(const char *Pattern) {
	int Type = IP_Pattern_Valid(Pattern);
	struct Pattern *Parsed;
	if (!Type) return -1;

	Parsed = Parse_Pattern(Pattern, Type);
	if (!Parsed->Valid) return Type;

	/* Pattern is really valid and already parsed (but stored statically) */
	PatternCount++;
	PatternMem = realloc(PatternMem, sizeof(struct Pattern*) * PatternCount);
	PatternMem[PatternCount - 1] = CopyPattern(Parsed);
	
	return 0;
}

