/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <libintl.h>
#include <sys/types.h>

#include "ProcessTable.h"
#include "MutualExclusion.h"
#include "Tools.h"
#include "sambascanner.h"

extern struct Options_Structure Options;
static struct Process *Processes = NULL;
static struct Process *CurrentProcess = NULL;
static int NumberProcesses = 0;

#ifdef DEBUG
static int ProcessTable_CheckIntegrity(char AbortOnError) {
	struct Process *Current;
	Lock_ProcessTable();
	Current = Processes;
	while (Current != NULL) {
		if (Current->Next != NULL) {
			if (Current->Next->Prev != Current) {
				if (AbortOnError) {
					fprintf(stderr, _("ProcessTable corruption detected. Aborting.\n"));
					abort();
				}
				return 0;
			}
		}
		Current = Current->Next;
	}
	Unlock_ProcessTable();
	return 1;
}

static void ProcessTable_Display() {
	struct Process *Current = Processes;
	ProcessTable_CheckIntegrity(1);
	Lock_ProcessTable();
	fprintf(stderr, "PTBL: ");
	while (Current != NULL) {
		fprintf(stderr, "[ %p, %d %d ] -> ", Current, Current->PID, Current->Running);
		Current = Current->Next;
	}
	fprintf(stderr, "NULL\n");
	Unlock_ProcessTable();
}
#endif

/* After rewinding the ProcessTable it has to be completely traversed using
 * ProcessTable_Next. The ProcessTable gets locked upon rewinding and gets
 * unlocked after there is no "Next" element */
void ProcessTable_Rewind() {
#ifdef DEBUG
	ProcessTable_CheckIntegrity(1);
	ProcessTable_Display();
#endif
	Lock_ProcessTable();
	CurrentProcess = NULL;
}

struct Process* ProcessTable_Next() {
	struct Process *CurrentProcessCopy;
	if (CurrentProcess == NULL) {
		/* List was just rewound */
		CurrentProcess = Processes;
	} else {
		CurrentProcess = CurrentProcess->Next;
	}
	CurrentProcessCopy = CurrentProcess;
	if (CurrentProcess == NULL) {
		Unlock_ProcessTable();
	}
	return CurrentProcessCopy;
}

struct Process* ProcessTable_CreatePreliminary(double TermTime, double KillTime) {
	struct Process *New;
	double CurrentTime = GetTime();

	Lock_ProcessTable();

	New = (struct Process*)malloc(sizeof(struct Process));
	if (New == NULL) {
		perror("malloc ProcessTable_CreatePreliminary");
		exit(EXIT_FAILURE);
	}
	
	NumberProcesses++;

	New->Running = 1;
	New->Status = 0;
	New->Forwarding = 1;
	New->Signalling_State = 0;
	New->TermTime = CurrentTime + (TermTime * Options.Timings);
	New->KillTime = CurrentTime + ((TermTime + KillTime) * Options.Timings);
	New->PID = 0;

	New->Next = Processes;
	New->Prev = NULL;
	if (Processes != NULL) Processes->Prev = New;
	Processes = New;
	
	CurrentProcess = New;
	return New;
}

void ProcessTable_UpdateEntry_ThreadID(pthread_t ThreadID) {
	assert(CurrentProcess != NULL);
	CurrentProcess->ThreadID = ThreadID;
}

void ProcessTable_FinishEntry(pid_t PID) {
	assert(CurrentProcess != NULL);
	CurrentProcess->PID = PID;
	Unlock_ProcessTable();
}

void ProcessTable_UpdateEntry_Times(struct Process *Proc, double TermTime, double KillTime) {
	double CurrentTime = GetTime();
	Lock_ProcessTable();
	Proc->TermTime = CurrentTime + (TermTime * Options.Timings);
	Proc->KillTime = CurrentTime + ((TermTime + KillTime) * Options.Timings);
	Unlock_ProcessTable();
}

pid_t ProcessTable_GetCurrent_PID() {
	assert(CurrentProcess != NULL);
	return CurrentProcess->PID;
}

char ProcessTable_GetCurrent_Running() {
	assert(CurrentProcess != NULL);
	return CurrentProcess->Running;
}

double ProcessTable_GetCurrent_TermTime() {
	assert(CurrentProcess != NULL);
	return CurrentProcess->TermTime;
}

double ProcessTable_GetCurrent_KillTime() {
	assert(CurrentProcess != NULL);
	return CurrentProcess->KillTime;
}

int ProcessTable_GetCurrent_SignallingState() {
	assert(CurrentProcess != NULL);
	return CurrentProcess->Signalling_State;
}

char ProcessTable_Get_Running(struct Process *Proc) {
	char Running;	
	assert(Proc != NULL);
	Lock_ProcessTable();
	Running = Proc->Running;
	Unlock_ProcessTable();
	return Running;
}

pid_t ProcessTable_Get_PID(struct Process *Proc) {
	pid_t PID;	
	assert(Proc != NULL);
	Lock_ProcessTable();
	PID = Proc->PID;
	Unlock_ProcessTable();
	return PID;
}

pthread_t ProcessTable_Get_TID(struct Process *Proc) {
	pthread_t TID;	
	assert(Proc != NULL);
	Lock_ProcessTable();
	TID = Proc->ThreadID;
	Unlock_ProcessTable();
	return TID;
}

int ProcessTable_Get_Status(struct Process *Proc) {
	int Status;	
	assert(Proc != NULL);
	Lock_ProcessTable();
	Status = Proc->Status;
	Unlock_ProcessTable();
	return Status;
}

int ProcessTable_Get_Forwarding(struct Process *Proc) {
	int Fwd;
	assert(Proc != NULL);
	Lock_ProcessTable();
	Fwd = Proc->Forwarding;
	Unlock_ProcessTable();
	return Fwd;
}

void ProcessTable_SetCurrent_NotRunning() {
	assert(CurrentProcess != NULL);
	CurrentProcess->Running = 0;
}

void ProcessTable_SetCurrent_Status(int Status) {
	assert(CurrentProcess != NULL);
	CurrentProcess->Status = Status;
}

void ProcessTable_SetCurrent_SignallingState(int State) {
	assert(CurrentProcess != NULL);
	CurrentProcess->Signalling_State = State;
}

void ProcessTable_Set_Forwarding_Closed(struct Process *Proc) {
	Lock_ProcessTable();
	Proc->Forwarding = 0;
	Unlock_ProcessTable();
}

/* RemoveEntry can be called from external source */
void ProcessTable_RemoveEntry(struct Process *ToRemove) {
	Lock_ProcessTable();
	if (ToRemove->Prev != NULL) ToRemove->Prev->Next = ToRemove->Next;
	if (ToRemove->Next != NULL) ToRemove->Next->Prev = ToRemove->Prev;
	if (ToRemove == Processes) Processes = Processes->Next;
	free(ToRemove);
	Unlock_ProcessTable();
}

/* FreeCurrent can be called while traversal of process list */
static void ProcessTable_FreeCurrent() {
	struct Process *NewCurrent = NULL;
	if (CurrentProcess->Prev != NULL) {
		CurrentProcess->Prev->Next = CurrentProcess->Next;
		NewCurrent = CurrentProcess->Prev;
	}
	if (CurrentProcess->Next != NULL) {
		CurrentProcess->Next->Prev = CurrentProcess->Prev;
	}
	if (CurrentProcess == Processes) {
		Processes = Processes->Next;
	}
	free(CurrentProcess);
	CurrentProcess = NewCurrent;
}

void ProcessTable_Free() {
	ProcessTable_Rewind();
	while (ProcessTable_Next()) ProcessTable_FreeCurrent();
}

