/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005-2007 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include "config.h"

#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <libintl.h>

#include "PThreads.h"
#include "String.h"
#include "sambascanner.h"
#include "smblister.h"

void Init_StringList(StringList *List) {
	assert(List != NULL);
	List->Entries = 0;
	List->Entry = NULL;
	Safe_pthread_mutex_init(&(List->Mutex), "StringList Mutex");
}

void Destroy_StringList(StringList *List) {
	int i;
	assert(List != NULL);
	for (i = 0; i < List->Entries; i++) free(List->Entry[i]);
	free(List->Entry);
	List->Entries = 0;
	List->Entry = NULL;
	Safe_pthread_mutex_destroy(&(List->Mutex), "StringList Mutex");
}

void Remove_Trailing_Spaces(char *String) {
	int i, l;
	assert(String != NULL);
	l = strlen(String);
	for (i = l - 2; i >= 0; i--) {			/* TODO: Is das 'l - 2' hier ein Bug? Müßte das nicht 'l - 1' heißen? */
		if (String[i] == ' ') {
			String[i] = 0;
		} else {
			break;
		}
	}
}

void Remove_Trailing_Whitespaces(char *String) {
	int i, l;
	assert(String != NULL);
	l = strlen(String);
	for (i = l - 1; i >= 0; i--) {
		if ((String[i] == ' ') || (String[i] == '\t')) {
			String[i] = 0;
		} else {
			break;
		}
	}
}

void Remove_Leading_Whitespaces(char *String) {
	int i, l, c;
	assert(String != NULL);
	l = strlen(String);
	c = -1;
	for (i = 0; i < l; i++) {
		if ((String[i] != ' ') && (String[i] != '\t')) break;
	}
	c = i;
	if (c == 0) return;

	for (i = 0; i < l - c; i++) {
		String[i] = String[i + c];
	}
	String[l - c] = 0;
}

int String_Exists(const StringList *List, const char *String) {
	int i;
	assert(List != NULL);
	assert(String != NULL);
	for (i = 0; i < List->Entries; i++) {
		if (!strcasecmp(List->Entry[i], String)) return i;
	}
	return -1;
}

void Lock_Stringlist(StringList *List) {
	assert(List != NULL);
	pthread_mutex_lock(&(List->Mutex));
}

void Unlock_Stringlist(StringList *List) {
	assert(List != NULL);
	pthread_mutex_unlock(&(List->Mutex));
}

void Add_String(StringList *List, const char *String) {
	assert(List != NULL);
	assert(String != NULL);
	List->Entry = (char**)realloc(List->Entry, sizeof(char**) * (List->Entries + 1));
	List->Entry[List->Entries] = (char*)malloc(sizeof(char) * (strlen(String) + 1));
	strcpy(List->Entry[List->Entries], String);
	List->Entries++;
}

void Add_String_Atomic(StringList *List, const char *String) {
	assert(List != NULL);
	assert(String != NULL);
	Lock_Stringlist(List);
	Add_String(List, String);
	Unlock_Stringlist(List);
}

void Add_String_Unique(StringList *List, const char *String) {
	assert(List != NULL);
	assert(String != NULL);
	if (String_Exists(List, String) != -1) return;
	Add_String(List, String);
}

void Add_String_Unique_Atomic(StringList *List, const char *String) {
	assert(List != NULL);
	assert(String != NULL);
	Lock_Stringlist(List);
	Add_String_Unique(List, String);
	Unlock_Stringlist(List);
}

void Remove_String(StringList *List, int Index) {
	assert(List != NULL);
	/*
	assert(Index >= 0);
	assert(Index < List->Entries);
	*/
	if ((Index < 0) || (Index >= List->Entries)) {
		fprintf(stderr, _("Fatal error: Removal of StringList index %d out of bounds. Have %d entries total.\n"), Index, List->Entries);
		exit(EXIT_FAILURE);
	}
	Lock_Stringlist(List);
	if (List->Entries == 1) {
		/* Remove the one and only entry in Stringlist */
		free(List->Entry[0]);
		free(List->Entry);
		List->Entries = 0;
		List->Entry = NULL;
	} else {
		free(List->Entry[Index]);
		if (List->Entries - 1 != Index) {
			/* Remove the not (by index) last entry in Stringlist */
			List->Entry[Index] = List->Entry[List->Entries - 1];
		}
		List->Entries--;
		List->Entry = (char**)realloc(List->Entry, sizeof(char**) * List->Entries);
	}
	Unlock_Stringlist(List);
}

void String_Lowercase(char *String) {
	int i;
	int l;
	assert(String != NULL);
	l = strlen(String);
	for (i = 0; i < l; i++) {
		if ((String[i] >= 'A') && (String[i] <= 'Z')) {
			String[i] -= 'A' - 'a';
		}
	}
}

const char *SingularPlural(const char *Singular, const char *Plural, int Number) {
	return (Number == 1) ? Singular : Plural;
}

void Date_To_String(char *Buffer, int Size, time_t Time) {
	struct tm Tm;
	assert(Buffer != NULL);
	assert(Size >= 20);
	if (localtime_r(&Time, &Tm) == NULL) {
		fprintf(stderr, "Error converting date '%ld': %s\n", Time, strerror(errno));
		Buffer[0] = 0;
		return;
	}

	snprintf(Buffer, Size, "%04d-%02d-%02d-%02d-%02d-%02d", Tm.tm_year + 1900, Tm.tm_mon + 1, Tm.tm_mday, Tm.tm_hour, Tm.tm_min, Tm.tm_sec);
}

const char *GetSmblisterError(int Status) {
	switch (Status) {
		case RETURN_SUCCESS:				return _("Success");
		case RETURN_SYNTAX:					return _("Syntax error");
		case RETURN_UNKNOWN_PARAMETER:		return _("Unknown parameter");
		case RETURN_PARAMETERS_USERPASS:	return _("Too little parameters for user/pass");
		case RETURN_INVALID_USERPASS1:		return _("Invalid User/Password #1");
		case RETURN_INVALID_USERPASS2:		return _("Invalid User/Password #2");
		case RETURN_SMBC_FAILURE:			return _("Failed to init libsmbclient");
		case RETURN_PASSWORDPROTECTED:		return _("Share is password protected");
	}
	return _("Unknown smblister return code");
}

